// CS350
// Eric Bengfort
// Proj III
// ebengfor@digipen.edu
// Due 03-21-07

////////////////////////////////////////////////////////////////////////
// $Id: render.cpp 71 2005-04-11 21:09:15Z gherron $
//
// Here is where the scene is drawn.  All student implementation will
// be in this file.
//
// Copyright  2005 DigiPen Institute of Technology
////////////////////////////////////////////////////////////////////////////////

#include "framework.h"
#include "scene.h"
#include <assert.h> // for assert.
#include <algorithm> // for sort.
#include <math.h>

// The width and height of the window during the last/previous call to "DrawScene()"
// If these values have since changed, must free and allocate a new Z-Buffer and
// ColorBuffer.  They are obviously globals.
int g_prevHeight = -1;
int g_prevWidth = -1;

// These are globals, just allocate "something here" and deal with it below.
float *zBuffer = new float[5];  // Allocated!
Color *colorBuffer = new Color[5];  // Allocated!

/*******************************************
List of Terms and various notes to myself:
Edge Table (ET) [perhaps a linked list?].  ET is called a bucket.

AEL: Active Edge List, list of edges crossing scanline.
AEL often will not change (hence the efficiency)

Each Edge has x0 y0 z0 and x1 y1 z1

EXE has no clipper, make sure pixels are on the screen.
We'll lose credit if drag off screen == crash.

Origin is bottom left.

This algorithm takes 1 entire polygon at a time and
does not break it into triangles.

This algorithm goes "up" so to speak.  Y increases,
move up screen.

Either pretend scanlines & pixels go between half intervals or do it another way.
Either way, decide what range of values a scanline occurs.

My lecture notes state that Herron likes scanlines representing one integer to the next.

Given a float, what is the largest scanline less than or equal to it (a ceil or a floor)

Conversion to Pixel Space:
Xs = ( (Xp + 1) / 2 ) * w = (w/2) * Xp  +  w/2
Ys = ( (Yp + 1) / 2 ) * h = (h/2) * Yp  +  h/2
Zs = Zp

ex and ez seem to be current position as moving along.

********************************************/


// vvv use Color Class INSTEAD!
//class ColorNode // vals live [0,1]
//{
//	float mRed;
//	float mGreen;
//	float mBlue;
//};

//-float g_previous_d = 1.7f; // just start it somewhere reasonable.

// Make an instance of this class for each unique texture map one time.
// Save em into gMIPMAPS.
class MipMap
{
public:
	// The levels zero to ___. (probably 0 to 9) [if 256x256]
	// Level Zero is regular texture map
	// Level One is average of above finer level 
	// [each color here is average of above 2x2]
	std::vector< std::vector<Color> > m_MIP;
	Texture *m_pTM; // ptr to TM which this MM was made from.
	int m_width; // width of level zero of this MM
	int m_height; // height of level zero of this MM
	int m_levels; // total number of levels of this MM.

	MipMap(int w, int h, Texture *TM) // call with w and h of TM itself and ptr to TM itself to make MM from.
	{
		assert(w == h);
		assert(w == 256);
		assert(h == 256);
		assert(TM != NULL);
		
		m_pTM = TM;
		m_width = w;
		m_height = h;

		// Allocate array space.
		// Find # levels first.  NUMBER OF.
		int temp = w;
		m_levels = 1;
		while (temp > 1)
		{
			temp = temp / 2;
			++m_levels;
		}
		assert(temp == 1); // make sure all this division by 2 leads here, else OH NO!

		std::vector<Color> aLevel;
		for (int a = 0; a < m_levels; ++a)
		{
			aLevel.clear();
			for (int b = 0; b < w * h; ++b)
			{
				aLevel.push_back( Color() ); // pushback nothing of real value, give value later.
			}
			m_MIP.push_back(aLevel);
			w = w / 2;
			h = h / 2;
		}
		assert(m_MIP.size() == m_levels);

		// Okay, NOW fill in Level 0 of MM.  Then make following levels based upon LV 0.

		for (int c = 0; c < (int)m_MIP[0].size(); ++c)
		{
			// convert each 'c' into a u,v [like an x,y] to index into orig TM

			int i = c;
			int t = 0; // for how many times tOOK away.
			while (i > m_width)
			{
				i -= m_width;
				++t;
			}
			// (x,y) = (new i, # times took away)

			const int uu = i;
			const int vv = t;

			// This COULD be peril, needing to go to one less than width/height possible
			// instead of less than or equal to.
			assert( uu >= 0 && uu <= TM->width);
			assert( vv >= 0 && vv <= TM->height);

			unsigned char *r = TM->byte(uu, vv, 0);
			unsigned char *g = TM->byte(uu, vv, 1);
			unsigned char *b = TM->byte(uu, vv, 2);
			
			float red = (float)(*r) / 255.f;
			float green = (float)(*g) / 255.f;
			float blue = (float)(*b) / 255.f;
			assert(red >= 0.f && red <= 255.f);
			assert(green >= 0.f && green <= 255.f);
			assert(blue >= 0.f && blue <= 255.f);
			//-psColor.Set(red, green, blue, 1);
			m_MIP[0][c].Set(red, green, blue, 1);
		}

		// neo4, just need 1st level, so leave! [aka lv 0]
		return;

		// Now make remaining levels...
		// reuse w and h
		w = m_width;
		h = m_height;
		int numLevelsLeftToMake = m_levels - 1;
		int makeMe = 1;
		while (numLevelsLeftToMake > 0)
		{
			LevelBelowMaker(makeMe, w, h); // will be called last time with w=h=2 since uses LEVEL ABOVE'S STATS to make the current level [which in this case would be w=h=1]


			w = w / 2;
			h = h / 2;
			++makeMe;
			--numLevelsLeftToMake;
		}
		assert(h == 1 && w == 1);
		assert(makeMe == 9); // should be 8 but ++ goes off.

		
	} // end of constructor FN.

	void LevelBelowMaker(int makeLevelID,  // level X+1, based upon avging from Level X.
						 int widthLevelX,
						 int heightLevelX )
	{
		// So, we call this FN with makeLevelID 1, 2, 3, ... [level 0 already made]
		// fromLevelID = makeLevelID - 1.
		// width and height of LevelX are from 0, 1, 2, ...
		int fromLevelID = makeLevelID - 1;

		int j = 0; // index into level I'm making for avg.

		//**int debugTestStar = 0;

		int c = 0;
		for (int i = 0; i < widthLevelX * heightLevelX; i += 2) // loop thru Level X
		{
			// Do it's 3 neighbors.
			assert(makeLevelID == fromLevelID + 1);
			m_MIP[makeLevelID][j] = GetAverage(i, widthLevelX, fromLevelID);
			//**++debugTestStar;
			++c;
			if (c == widthLevelX/2)
			{
				c = 0;
				i = i + widthLevelX;
			}
			++j;
		}
		//**assert(debugTestStar == (heightLevelX * widthLevelX) / 4);

	} // end of FN,

	// given an index i, finds avg of itself and 3 nearby neighbors and rets as the avg'd color.
	Color GetAverage(int i, int width, int level) // takes index and width of a MM, level above the one we're making.
	{
		Color color;
		//this->m_MIP[level][i].rgba
		for (int b = 0; b < 3; ++b)
		{
			color.rgba[b] = (m_MIP[level][i].rgba[b] + m_MIP[level][i+1].rgba[b] + m_MIP[level][i+width].rgba[b] + m_MIP[level][i+width+1].rgba[b]) / 4.f;
		}
		color.rgba[3] = 1.f;

		return color;
	}

};


std::vector<MipMap> gMIPMAPS; // global vec of ALL mipmaps made for each TM for each object!


// Class holding all the data an edge needs to know.
class EdgeNode
{
public:

	float x0;	// X Component of point representing the starting point of the edge (which is a line segment)
	float y0;	// Y Component of point representing the starting point of the edge (which is a line segment)
	float z0;	// Z Component of point representing the starting point of the edge (which is a line segment)
	float x1;	// X Component of point representing the ending point of the edge (which is a line segment)
	float y1;	// Y Component of point representing the ending point of the edge (which is a line segment)
	float z1;	// Z Component of point representing the ending point of the edge (which is a line segment)
			 
	float dxdy;	// Change in x with respect to y.
	float dzdy;	// Change in z with respect to y.
	//Color edgeColor; // Quick Lighted Color with simple dot product method. [never used/saved]

	Vector3D N0; // neo
	Vector3D N1; // neo
	Vector3D dNdy; // neo

	Vector3D T0; // neo2 (u', v', w')
	Vector3D T1; // neo2 (u', v', w')
	Vector3D dTdy; // neo2

	bool removeMe; // Flag to remove a node from a list.

	// neo2 removed const for verticies
	EdgeNode(Vertex &start, Vertex &end, Object &obj, float startW, float endW)
	{
		// neo2
		assert(start.u >= 0);
		assert(start.u <= 1);
		assert(start.v >= 0);
		assert(start.v <= 1);

		assert(end.u >= 0);
		assert(end.u <= 1);
		assert(end.v >= 0);
		assert(end.v <= 1);
		// neo2

		// u and v live in [0,1]


		removeMe = false; // When time to be removed, set to true.  A flag.
		// Setup Values for the data that an edge needs to know.
		//x0 = start.xyz[0];
		//y0 = start.xyz[1];
		//z0 = start.xyz[2];
		//x1 = end.xyz[0];
		//y1 = end.xyz[1];
		//z1 = end.xyz[2];
		// ^^^ NOT VECTORS ANYMORE!

		// Okay, despite their names start & end, I might not call properly with
		// the 2 verticies in the proper order.  So, deal with that here and now!
		if (start.V.xyzw[1] <= end.V.xyzw[1])
		{
			// x0 to z1 are member data of the EdgeNode Class.  We're filling them in.
			x0 = start.V.xyzw[0];
			y0 = start.V.xyzw[1];
			z0 = start.V.xyzw[2];
			N0 = start.N;  // neo
			T0 = Vector3D( start.u / startW, start.v / startW, 1 / startW ); // neo2
			x1 = end.V.xyzw[0];
			y1 = end.V.xyzw[1];
			z1 = end.V.xyzw[2];
			N1 = end.N; // neo
			T1 = Vector3D( end.u / endW, end.v / endW, 1 / endW ); // neo2
		}
		else
		{
			// x0 to z1 are member data of the EdgeNode Class.  We're filling them in.
			x0 = end.V.xyzw[0];
			y0 = end.V.xyzw[1];
			z0 = end.V.xyzw[2];
			N0 = end.N;  // neo
			T0 = Vector3D( end.u / endW, end.v / endW, 1 / endW ); // neo2
			x1 = start.V.xyzw[0];
			y1 = start.V.xyzw[1];
			z1 = start.V.xyzw[2];
			N1 = start.N;  // neo
			T1 = Vector3D( start.u / startW, start.v / startW, 1 / startW ); // neo2
		}

	
		const float dy = y1 - y0; // change in y, NOT saved within class data members.
		
		dxdy = (x1 - x0) / dy;
		dzdy = (z1 - z0) / dy;

		// neo //
		dNdy[0] = (N1[0] - N0[0]) / dy;
		dNdy[1] = (N1[1] - N0[1]) / dy;
		dNdy[2] = (N1[2] - N0[2]) / dy;
		// neo //

		// neo 2 //
		dTdy[0] = (T1[0] - T0[0]) / dy;
		dTdy[1] = (T1[1] - T0[1]) / dy;
		dTdy[2] = (T1[2] - T0[2]) / dy;
		// neo 2 //

		// Okay, we've almost got it.  Last Peril, we might be off-screen.
		if (y0 < 0.f)
		{
			// okay, just move it to the beginning (line zero) and update x and y
			// for the new defined location.  This was my explaination.

			const float distToScanLineZero = 0.f - y0;  // above pt's y pos - below pt's y pos.
			assert(distToScanLineZero > 0.f); // think about it, y0 is neg.  zero minus a neg == zero plus a pos => pos.
			x0 = x0 + (distToScanLineZero * dxdy);
			y0 = 0.f;
			z0 = z0 + (distToScanLineZero * dzdy);
			// neo //
			N0[0] = N0[0] + (distToScanLineZero * dNdy[0]);
			N0[1] = N0[1] + (distToScanLineZero * dNdy[1]);
			N0[2] = N0[2] + (distToScanLineZero * dNdy[2]);
			// neo //
			// neo2 //
			T0[0] = T0[0] + (distToScanLineZero * dTdy[0]);
			T0[1] = T0[1] + (distToScanLineZero * dTdy[1]);
			T0[2] = T0[2] + (distToScanLineZero * dTdy[2]);
			// neo2 //
		}

		assert(y0 >= 0.f); // make sure we're visible / on-screen at the end of the day.

	} // end of my constructor function.

}; // end of my Node class.


// For sorting EdgeNodes by x0 component.
bool SortAELbyX(const EdgeNode &a, const EdgeNode &b)
{
	if (a.x0 <= b.x0)
	{
		return true;
	}
	else
	{
		return false;
	}
}



// Used to switch openGL polygon rendering
bool useOpenGLRendering = true;
bool useFastRendering = false;

// Given Two 3D Vectors, returns the cross product of those two vectors (obviously).
Vector3D CrossProduct(const Vector3D &a, const Vector3D &v)
{
	Vector3D res;
	res.xyz[0] = a.xyz[1] * v.xyz[2] - a.xyz[2] * v.xyz[1];
	res.xyz[1] = a.xyz[2] * v.xyz[0] - a.xyz[0] * v.xyz[2];
	res.xyz[2] = a.xyz[0] * v.xyz[1] - a.xyz[1] * v.xyz[0];
	return res;
}

// returns by reference a bumped version of N_terpolate.  
// Use this returned value for phong lighting.
// Make sure to NORMALIZE him before use.
void BumpTheNormal(Vector3D &N_terpolate, Scene& scene, Vector3D &T_terpolate, Object &obj, Vector3D &dTdx, Vector3D &dTdy)
{
	// Nbar = N + D [all 3 are vectors]
	// N_terpolate = N_terpolate + D [this is what I shall do]

	// To calculate D we need the following:
	/****************
	* N From the scanline algorithm [got it, is N_terpolate]
	* fu and fv, these come from Bump Map.  Come from differences of (adjacent) 
		values in the bump map.
	* Pu  Pv [got em, just (1,0,0) and (0,1,0) respectively.
	****************/

	/*******************
	From Lecture Notes: Day 03-14-07 Page One

	D = ( FvPu - FuPv ) X ( N / ||N|| )
	Fu = @F/@u
	Fv = @F/@v
	Pu = @P/@u
	Pv = @P/@v

	*******************/

	/************************
	[ this works ]
	Vector3D a(1,0,0);
	Vector3D b(0,1,0);
	Vector3D c = CrossProduct(a, b);
	*************************/

	assert(obj.bump != NULL);
	int w = obj.bump->width;
	int h = obj.bump->height;

	float u = T_terpolate[0] / T_terpolate[2];
	float v = T_terpolate[1] / T_terpolate[2];

	if (u < 0.f)
		u = 0.f;
	if (u > 1.f)
		u = 1.f;
	if (v < 0.f)
		v = 0.f;
	if (v > 1.f)
		v = 1.f;

	//--------------------------

	int uu = (int)( u * (  (float)w  ) );
	int vv = (int)( v * (  (float)h ) );


	Vector3D dTTdx = dTdx;
	Vector3D dTTdy = dTdy;
	dTTdx[0] = dTTdx[0] / T_terpolate[2];
	dTTdx[1] = dTTdx[1] / T_terpolate[2];
	dTTdy[0] = dTTdy[0] / T_terpolate[2];
	dTTdy[1] = dTTdy[1] / T_terpolate[2];


	int uu_plus_1 = uu + dTTdx[0];
	int vv_plus_1 = vv + dTTdy[1];

	if (uu_plus_1 >= 255)
	{
		uu_plus_1 = 255;
	}
	if (vv_plus_1 >= 255)
	{
		vv_plus_1 = 255;
	}

	if (uu_plus_1 <= 0)
	{
		uu_plus_1 = 0;
	}
	if (vv_plus_1 <= 0)
	{
		vv_plus_1 = 0;
	}

	//--------------------------

	// unsigned char *r = TM->byte(uu, vv, 0);
	// Probably don't need i, o, and p at all.  Use uu and vv equivilents
	// living between 0 and w or h respectively.

	// Fu = F(u+1, v) - F(u, v) / (u+1 - u)
	// Fv = F(u, v+1) - F(u, v) / (v+1 - v)


	float h_uv = (float)(*(obj.bump->byte(uu, vv, 0)));
	float h_u1v = (float)(*(obj.bump->byte(uu_plus_1, vv, 0)));
	float h_uv1 = (float)(*(obj.bump->byte(uu, vv_plus_1, 0)));

	float Fu = (h_u1v - h_uv) / w;
	float Fv = (h_uv1 - h_uv) / h;

    //float Fu = obj.bump->byte(uu_plus_1, vv, 0) - obj.bump->byte(uu, vv, 0)

	Vector3D tempNorm = N_terpolate;
	//tempNorm = (1.f / tempNorm.Normalize()) * N_terpolate;
	tempNorm.Normalize(); // 2nd piece of CP
	Vector3D Pu(1,0,0);
	Vector3D Pv(0,1,0);
	//Vector3D firstPiece = (Fv * Pu) - (Fu * Pv);
	Vector3D firstPiece;
	firstPiece[0] = -Fv * Pu[0] + Fu * Pv[0];
	firstPiece[1] = -Fv * Pu[1] + Fu * Pv[1];
	firstPiece[2] = -Fv * Pu[2] + Fu * Pv[2];
	Vector3D D = CrossProduct( firstPiece, tempNorm );
	N_terpolate = N_terpolate + D;

} // end of function!

void MipMapTextureMagicGo(Color &psColor, Scene& scene, Vector3D &T_terpolate, Object &obj, Vector3D &dTdx, Vector3D &dTdy)
{
	assert(obj.texture != NULL);

	float u = T_terpolate[0] / T_terpolate[2];
	float v = T_terpolate[1] / T_terpolate[2];

	if (u < 0)
		u = 0;
	if (u > 1)
		u = 1;
	if (v < 0)
		v = 0;
	if (v > 1)
		v = 1;

	/////////////////////////////////




	// Okay, now need to find level to dive into
	// Will almost surely be BETWEEN two levels, so we'll make
	// uuA, vvA  and  uuB, vvB.  Dive into both and weighted avg em
	// for ONE color we then useth!


	// Changing the base of an algorithm:
	// log base a of x is equal to      log base b of x divided by log base b of a
	// IGNORE THIS LINE: therefore log base 2 of PixelArea is Log base ANYTHING of PixelArea DIVIDED by log base ANYTHING of

	// a = 2
	// x = pixel area
	// Log Base 2 of Pixel Area = Log Base ANY of Pixel Area DIVIDED BY Log Base ANY of 2.

	// STEP ONE:  Calculate A.
	// STEP TWO:  Calculate d.
	// STEP THREE:  Calculate f
	// STEP FOUR:  Delve into levels f and f+1 of MMs and blend the colors.
	// STEP FIVE:  Use that color.
	// STEP SIX:  Profit!

	/***
	We'll have 2 uu's and 2 vv's for each level.
	(we index into).  The u and the v are set in stone though
	Also, can't just use obj.texture, must use its MM!
	AND be weary of the real w/h at current level we are in!
	int uu = (int)( u * (obj.texture->width) );
	int vv = (int)( v * (obj.texture->height) );

	assert( uu >= 0 && uu <= obj.texture->width);
	assert( vv >= 0 && vv <= obj.texture->height);
	***/


	for (int k = 0; k < (int)gMIPMAPS.size(); ++k)
	{
		if (gMIPMAPS[k].m_pTM == obj.texture)
		{
			break; // use index k to get the right MM. [2 MM's probably exist]
		}
	}


	//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	const bool JustUseTextureMap = true; 
	// Set this to TRUE when shipping.  True turns OFF mipmapping and 
	// forces use of level 0, which is really just the texture map itself.

	if (JustUseTextureMap)
	{
		int f = 0;

		int uuF = (int)( u * (  (gMIPMAPS[k].m_width / pow(2,f))    ) );
		int vvF = (int)( v * (  (gMIPMAPS[k].m_height / pow(2,f))    ) );

		int iF = uuF + (vvF * (gMIPMAPS[k].m_width / pow(2,f)));

		int wwwF = (int)gMIPMAPS[k].m_MIP[f].size();
		if (iF < wwwF)
		{
			// okay
		}
		else
		{
			iF = wwwF - 1;
		}

		assert( iF < (int)gMIPMAPS[k].m_MIP[f].size() );

		psColor.rgba[0] = gMIPMAPS[k].m_MIP[f][iF].rgba[0];
		psColor.rgba[1] = gMIPMAPS[k].m_MIP[f][iF].rgba[1];
		psColor.rgba[2] = gMIPMAPS[k].m_MIP[f][iF].rgba[2];
		psColor.rgba[3] = 1;

		return; // our work here is done.
	}
	//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	Vector3D dTTdx = dTdx;
	Vector3D dTTdy = dTdy;
	dTTdx[0] = dTTdx[0] / T_terpolate[2];
	dTTdx[1] = dTTdx[1] / T_terpolate[2];
	dTTdy[0] = dTTdy[0] / T_terpolate[2];
	dTTdy[1] = dTTdy[1] / T_terpolate[2];

	float PixelArea = (dTTdx[0] * dTTdy[1]) - (dTTdy[0] * dTTdx[1]);

	if (PixelArea < 0.f)
	{
		PixelArea = -PixelArea;
	}

	float d; // filled in below 

	{
		// assertion fired once: assert(PixelArea >= 0.f);
		//~!~assert(PixelArea <= 1.f); // because without 256^2 aka L*W factor should be much less than one.
		PixelArea = PixelArea * gMIPMAPS[k].m_height * gMIPMAPS[k].m_width;
		//~!~PixelArea = PixelArea; // extra scaling factor.

		float logBaseEofPixelArea = log(PixelArea);
		float logBaseEof2 = log(2.f);
		d = (logBaseEofPixelArea / logBaseEof2) / 2.f;

		// Super safeguard
		if (d < 0.f)
		{
			d = 0.f;
		}
		
		// One more Super Safeguard
		if (d > 7.98f)
		{
			d = 7.98f;
		}

	}


	// Note to self:  Be careful if algorithm follows levels 0 to X or X to 0.
	// Don't want to be flip flopped.

	int f = (int)floor(d);
	Color ColorF; // color from indexing into level f
	Color ColorF1; // color from indexing into level f + 1


	// HOW TO:
	// Remember, get mipmap ID k!
	//int uuF = (int)( u * (  width at level f   ) );
	//int vvF = (int)( v * (  height at level f   ) );
	//int uuF1 = (int)( u * (  width at level f+1   ) );
	//int vvF1 = (int)( v * (  height at level f+1   ) );

	int uuF = (int)( u * (  (gMIPMAPS[k].m_width / pow(2,f))    ) );
	int vvF = (int)( v * (  (gMIPMAPS[k].m_height / pow(2,f))    ) );
	int uuF1 = (int)( u * (  (gMIPMAPS[k].m_width / pow(2,f+1))    ) );
	int vvF1 = (int)( v * (  (gMIPMAPS[k].m_height / pow(2,f+1))    ) );

	// These uu's and vv's should be analogous to earlier single uu and vv.
	// They are integers living between 0 and the width of the MM level they
	// came from.

	int iF = uuF + (vvF * (gMIPMAPS[k].m_width / pow(2,f)));
	int iF1 = uuF1 + (vvF1 * (gMIPMAPS[k].m_width / pow(2,f+1)));

	int wwwF = (int)gMIPMAPS[k].m_MIP[f].size();
	if (iF < wwwF)
	{
		// okay
	}
	else
	{
		iF = wwwF - 1;
	}

	int wwwF1 = (int)gMIPMAPS[k].m_MIP[f+1].size();
	if (iF1 < wwwF1)
	{
		// okay
	}
	else
	{
		iF1 = wwwF1 - 1;
	}


	//gMIPMAPS[k].m_MIP[0][uu,vv].rgba
	//  Found MM
	//                LV0
	//                     index into LV 0
	//                             Color at this index at Level 0.

	// turn uu,vv into a single index i into a 1D Array Thing.

	/*****************************
	// ZONE:
	int i = uu + (vv * gMIPMAPS[f].m_width);
	// Incompatiable Types : psColor = gMIPMAPS[f].m_MIP[0][i].rgba;

	int www = (int)gMIPMAPS[f].m_MIP[0].size();
	if (i < www)
	{
		// okay
	}
	else
	{
		i = www - 1;
	}
	****************/

	/***
	assert( i < (int)gMIPMAPS[f].m_MIP[0].size() );
	****/

	assert( iF < (int)gMIPMAPS[k].m_MIP[f].size() );
	assert( iF1 < (int)gMIPMAPS[k].m_MIP[f+1].size() );


	/***
	psColor.rgba[0] = gMIPMAPS[f].m_MIP[0][i].rgba[0];
	psColor.rgba[1] = gMIPMAPS[f].m_MIP[0][i].rgba[1];
	psColor.rgba[2] = gMIPMAPS[f].m_MIP[0][i].rgba[2];
	psColor.rgba[3] = 1;
	***/

	ColorF.rgba[0] = gMIPMAPS[k].m_MIP[f][iF].rgba[0];
	ColorF.rgba[1] = gMIPMAPS[k].m_MIP[f][iF].rgba[1];
	ColorF.rgba[2] = gMIPMAPS[k].m_MIP[f][iF].rgba[2];
	ColorF.rgba[3] = 1;

	ColorF1.rgba[0] = gMIPMAPS[k].m_MIP[f+1][iF1].rgba[0];
	ColorF1.rgba[1] = gMIPMAPS[k].m_MIP[f+1][iF1].rgba[1];
	ColorF1.rgba[2] = gMIPMAPS[k].m_MIP[f+1][iF1].rgba[2];
	ColorF1.rgba[3] = 1;

	//psColor.rgba[0] = (1 - d + f) * ColorF.rgba[0] + (d - f) * ColorF1.rgba[0];
	//psColor.rgba[1] = (1 - d + f) * ColorF.rgba[1] + (d - f) * ColorF1.rgba[1];
	//psColor.rgba[2] = (1 - d + f) * ColorF.rgba[2] + (d - f) * ColorF1.rgba[2];
	//psColor.rgba[3] = 1;

	float aaaa = 1.f - d + (float)(f);
	float bbbb = d - (float)(f);

	// now do lin interp of the 2 colors and use the made/blended color!

	psColor.rgba[0] = aaaa * ColorF.rgba[0] + bbbb * ColorF1.rgba[0];
	psColor.rgba[1] = aaaa * ColorF.rgba[1] + bbbb * ColorF1.rgba[1];
	psColor.rgba[2] = aaaa * ColorF.rgba[2] + bbbb * ColorF1.rgba[2];
	psColor.rgba[3] = 1;

	const bool debugColorTest = false; // make SURE is FALSE when shipping project.
	if (debugColorTest)
	{
		if (f == 0)
		{
			psColor.rgba[0] = 1.f;
			psColor.rgba[1] = 0.f;
			psColor.rgba[2] = 0.f;
			psColor.rgba[3] = 1;
		}

		else if (f == 1)
		{
			psColor.rgba[0] = 0.f;
			psColor.rgba[1] = 1.f;
			psColor.rgba[2] = 0.f;
			psColor.rgba[3] = 1;
		}

		else if (f == 2)
		{
			psColor.rgba[0] = 0.f;
			psColor.rgba[1] = 0.f;
			psColor.rgba[2] = 1.f;
			psColor.rgba[3] = 1;
		}

		else if (f == 3)
		{
			psColor.rgba[0] = 0.f;
			psColor.rgba[1] = 1.f;
			psColor.rgba[2] = 1.f;
			psColor.rgba[3] = 1;
		}

		else if (f >= 4)
		{
			psColor.rgba[0] = 1.f;
			psColor.rgba[1] = 0.f;
			psColor.rgba[2] = 1.f;
			psColor.rgba[3] = 1;
		}
	}



} // end of function...



// Do Phong Shading, returns through psColor the color for a single pixel.
// psColor : returned color [currently empty/null/undefined]
// objDiffuse : color of object itself (pixel really) [i.e. red, blue ,etc...]
// N_terpolate : Interpolated Normal for THIS pixel. [still in world cords]
// x,y, and z : positons of this pixel [in pixel space]
void PhongShading(Color &psColor, Color &objDiffuse, Scene& scene, Vector3D &N_terpolate, float x, float y, float z, int width, int height, Object &obj)
{
	// Remember, normalize vectors!!
	/****
	    The code to get the world space coordinates of the eye is
        S = Vector4D(0.0, 0.0, 0.0, 1.0))
        Vector3D Eye = scene.viewing.InverseTransform(S).Hdiv();
	****/
		Vector4D S = Vector4D(0.0, 0.0, 0.0, 1.0); // S is one of those 'temporary' variables.
		Vector3D EyePosition = scene.viewing.InverseTransform(S).Hdiv(); 
		// [Eye POSITION [not eye vector] in World]

	/****
    The code to transform a pixel coordinate point (x,y,z) to
    projection coordinates, then to world coordinates will look like
    this:
        S = Vector4D(x*2/width-1.0, y*2/height-1.0, z)
        Vector4D T = scene.projection.InverseTransform(S);
        Vector3D W = scene.viewing.InverseTransform(T).Hdiv();
	****/
        S = Vector4D(x*2/width-1.0, y*2/height-1.0, z);
        Vector4D T = scene.projection.InverseTransform(S);
        Vector3D P = scene.viewing.InverseTransform(T).Hdiv(); // was W, notes call it P
		// [P: Pixel in World, Pworld]

		Vector3D EyeVector = EyePosition - P;  // note to self: insure this is proper order.

		EyeVector.Normalize();
		N_terpolate.Normalize();

		// PHONG
		// I = Ie + IaKd + (Summation of Ii) * [ Kd(N dot Li) + Ks(Hi dot N')^n ]
		// Can substitute (H dot N)^n with (V dot R)^n or similarly.

		// Ie, Ia, Ii : light color [Ie emitted  Ii intensity of the ith light]
		// Kd, color
		
		//       Li + V
		// Hi = --------
		//      |Li + V|

		Vector3D ambientFactorVec;
		for (int f = 0; f < 3; ++f)
		{
			ambientFactorVec.xyz[f] = scene.ambient.rgba[f] * objDiffuse.rgba[f];
		}
		//-const int shininess = scene.currentShininess;
		const int n = obj.shininess;


		Vector3D sumLand(0,0,0);
		for (int i = 0; i < (int)scene.lights.size(); ++i) // loop thru all lights lighting The Scene
		{
			//scene.lights[i].P -- type Vector3D
			//scene.lights[i].C -- type Color
			Vector3D lightPos = scene.lights[i].P;
			Color lightColor = scene.lights[i].C;
			Vector3D lightVector = lightPos - P;  // note to self: insure this is proper order.
			lightVector.Normalize();
			float NdotLi = Dot(N_terpolate, lightVector);
			if (NdotLi < 0.f)
			{
				NdotLi = 0.f; // Dot Prod can be return negative scalar, would SUBTRACT from sum, BAD!
			}

			Vector3D A(objDiffuse.rgba[0] * NdotLi, objDiffuse.rgba[1] * NdotLi, objDiffuse.rgba[2] * NdotLi);
			Vector3D Hi = lightVector + EyeVector;
			Hi.Normalize();
			float HidotNraise_n = pow(  Dot(Hi, N_terpolate) , n ) ;
			Vector3D B(obj.specular.rgba[0] * HidotNraise_n, obj.specular.rgba[1] * HidotNraise_n, obj.specular.rgba[2] * HidotNraise_n);
			Vector3D C = A + B;
			Vector3D EndSum( lightColor.rgba[0] * C[0], lightColor.rgba[1] * C[1], lightColor.rgba[2] * C[2] );
			sumLand += EndSum;			
		} // end of looping through the lovely lights.

		psColor.rgba[0] = ambientFactorVec.xyz[0] + sumLand.xyz[0];
		psColor.rgba[1] = ambientFactorVec.xyz[1] + sumLand.xyz[1];
		psColor.rgba[2] = ambientFactorVec.xyz[2] + sumLand.xyz[2];

} // end of Phong Shading Function.


// Do ScanLine magic given a filled out and proper edgeList, a color, and W and H of window.
void ScanLine(std::vector<EdgeNode> &EdgeList,  /*Color color,*/ int width, int height, Scene& scene, Object &obj)
{
	// Note to self for debugging: Use 'EdgeList._Myfirst,5' in watch window.
	int removedNodes = 0;
	std::vector<EdgeNode> ActiveEdgeList;  // used later.
	std::vector<EdgeNode> ActiveEdgeListTemp;

	assert(ActiveEdgeList.size() == 0); // Better safe than sorry, 'somehow'
	// Okay, loop through all verticies in pixel space.
	// Make Edges out of them, and prep the edges properly.

	int minSL = 9999999;
	int maxSL = 0;

	// EdgeList is already created.  Loop through it and find max and min y values for upcoming loop.
	for (int w = 0; w < (int)EdgeList.size(); ++w)
	{
		if ((int)EdgeList[w].y0 < minSL)
		{
			minSL = (int)EdgeList[w].y0;
		}

		if ((int)EdgeList[w].y1 > maxSL)
		{
			maxSL = (int)EdgeList[w].y1; // <-- bug fixed, had y0 by mistake here.
		}
	}


	// Update: Sanity check maxSL and minSL
	if (maxSL > height)
	{
		maxSL = height;
	}

	if (minSL < 0)
	{
		minSL = 0;
	}

	
	// Go through all scanlines (of relevance)
	// ... then, find relevant edges.
	for (int y = minSL; y <= maxSL; ++y) // Scan-Line Iteration (for the win) [less than or equal to in this case]
	{

		// Step One.  Go through Active Edge List.  Get rid of any Edges we have
		// passed/advanced/are done with / have done.
		// Do this by flagging.
		// Lecture Notes Call This: "REMOVE PASSED EDGES FROM AEL"
		for (int i = 0; i < (int)ActiveEdgeList.size(); ++i)
		{
			// We are currently at scanline 'y'
			// An Edge goes from y0 to y1.
			// Therefore, if y1 is LESS than current scanline value y, remove it,
			// we're done with that particular edge!
			// Update: y1 is a float, cast it down!
			// Has to be either < or <=.  Make sure this is sane.
			// Okay, it is once again up to my 'consistency.'  Therefore, using
			// my current scheme, it is CAST DOWN and LESS THAN OR EQUAL TO.
			if (  (int)(ActiveEdgeList[i].y1) <= y)  
			{
				ActiveEdgeList[i].removeMe = true;
			}
		}

		// Step Two.  Actually get rid of the flagged RemoveMe Edges.
		ActiveEdgeListTemp.clear();
		for (int i = 0; i < (int)ActiveEdgeList.size(); ++i)
		{
			// Save the NON-removed edges into ActiveEdgeListTemp.
			if (  ActiveEdgeList[i].removeMe == false )
			{
				ActiveEdgeListTemp.push_back( ActiveEdgeList[i] );
			}
		}
		ActiveEdgeList.clear();
		ActiveEdgeList = ActiveEdgeListTemp;
		ActiveEdgeListTemp.clear();


		// Step Three (III)
		// Update the positions of where we are on the edges which are active.
		// Using the dxdy's, etc...
		// Lecture Notes Call This: "INCREMENT EDGES ON AEL"
		// 02-23-07 : We're going up 1 by y so use d_dy, yes!
		for (int i = 0; i < (int)ActiveEdgeList.size(); ++i)
		{
			ActiveEdgeList[i].x0 = ActiveEdgeList[i].x0 + ActiveEdgeList[i].dxdy;
			ActiveEdgeList[i].z0 = ActiveEdgeList[i].z0 + ActiveEdgeList[i].dzdy;
			// Okay, think for a second about what I have just done.
			// The (x0,y0,z0) can be thought of as 'where' I currently am on this line
			// making up an edge.  I just SHOVED the bottom endpoint [really current point] up a hair.
			// I do not directly manip y0 here.  That is like what 'y' / scanline is doing
			// on my behalf.

			// neo //
			ActiveEdgeList[i].N0 = ActiveEdgeList[i].N0 + ActiveEdgeList[i].dNdy;
			// neo //

			// neo2 //
			ActiveEdgeList[i].T0 = ActiveEdgeList[i].T0 + ActiveEdgeList[i].dTdy;
			// neo2 //
		}


		// Step Four (IV)
		// I'll just call this what the lecture notes call it.
		// Lecture Notes Call This: "MERGE NEW EDGES FROM ET[Y] (edgetable)" [I call mine EdgeList]
		// So, go see what new (if any) edges are NOW in play and add them to ActiveEdgeList.
		// Warning: Lecture Notes say to "Beware of EndPoint Duplicates!"
		bool GarbageDay = false; // if not outside of loop, gets reset and forgotten.
		for (int i = 0; i < (int)EdgeList.size(); ++i)
		{
			// Check for duplicate end points first!~!
			if( (int)(EdgeList[i].y0) == y)
			{
				// We have a desireable EdgeNode here where y0 is the proper value.
				// Scan/Search ActiveEdgeList for E.P.D.
				for (int j = 0; j < (int)ActiveEdgeList.size(); ++j)
				{
					if (EdgeList[i].x0 == ActiveEdgeList[j].x1 && EdgeList[i].y0 == ActiveEdgeList[j].y1 && EdgeList[i].z0 == ActiveEdgeList[j].z1)
					{
						// Duplicate Ahoy!
						ActiveEdgeList[j].removeMe = true;
						GarbageDay = true;
					}
				}
			}

			// Garbage-Day!
			if (GarbageDay)
			{
				ActiveEdgeListTemp.clear();
				for (int k = 0; k < (int)ActiveEdgeList.size(); ++k)
				{
					// Save the NON-removed edges into ActiveEdgeListTemp.
					if (  ActiveEdgeList[k].removeMe == false )
					{
						ActiveEdgeListTemp.push_back( ActiveEdgeList[k] );
					}
				}
				ActiveEdgeList.clear();
				ActiveEdgeList = ActiveEdgeListTemp;
				ActiveEdgeListTemp.clear();
			}
			// End Garbage Day!



			// Just be consistent with my plan.  Cast the y0's to the ints...
			if( (int)(EdgeList[i].y0) == y) // both type int now, will be equal.
			{
				ActiveEdgeList.push_back( EdgeList[i] ); // Add New Edge to ActiveEdgeList.  It begins at this scanline.
			}
		}



		// Step Five (V)
		// SORTED BY X-INTER
		// Okay, finally, ActiveEdgeList is sorted by x (in incr. order)
		// So, sort the entire thing! [by x0]
		sort(ActiveEdgeList.begin(), ActiveEdgeList.end(), SortAELbyX);


		// Change in z with respect to x.
		// This is done at the SCANLINE level mind you.
		//-const float dzdx;

		// FOR EACH PAIR E,F IN AEL:
			// COMPUTE Pz + dPz
			// FOR EACH PIXEL X BETWEEN E & F
				// IF Pz < ZBUFFER[X,Y]:
					// (RECORD) ZBUFFER[X,Y] = Pz
					// CBUFFER[X,Y] = COLOR OF P AT (X,Y)
				// Pz += dPz


		for (int m = 0; m < (int)ActiveEdgeList.size(); m = m + 2)
		{
			float dzdx = ( ActiveEdgeList[m+1].z0 - ActiveEdgeList[m].z0 ) / ( ActiveEdgeList[m+1].x0 - ActiveEdgeList[m].x0 );
			int e = (int)ActiveEdgeList[m].x0;
			int f = (int)ActiveEdgeList[m+1].x0;
			float e_z = ActiveEdgeList[m].z0; // Because going from e to f. FLOAT NOT INT!

			// neo //
			Vector3D N_terpolate = ActiveEdgeList[m].N0; // 3-15-07 don't want to change [m].N0 itself, needs to be (it's x val) left alone for next ++y incr.
			Vector3D dNdx = ( ActiveEdgeList[m+1].N0 - ActiveEdgeList[m].N0 ) / ( ActiveEdgeList[m+1].x0 - ActiveEdgeList[m].x0 );
			// neo //

			// neo2 //
			 // these start at a value and interpolate up to the value at the end point of this edge.
			Vector3D T_terpolate = ActiveEdgeList[m].T0;
			Vector3D dTdx = ( ActiveEdgeList[m+1].T0 - ActiveEdgeList[m].T0 ) / ( ActiveEdgeList[m+1].x0 - ActiveEdgeList[m].x0 );
			// neo2 //

			// Make sure e and f are valid, i.e. living in [0, width]

			if (e < 0)
			{
				// e is essentially x0, if not.
				// offset to proper starting position since was offscreen my friend.
				T_terpolate -= dTdx * e; // neo2
				N_terpolate -= dNdx * e; //neoneoneo
				e = 0;
			}
			if (e > width)
			{
				e = width;
			}
			if (f < 0)
			{
				f = 0;
			}
			if (f > width)
			{
				f = width;
			}

			for (int x = e; x < f; ++x)
			{
				if (zBuffer[x + (width*y)] > e_z)
				{
					zBuffer[x + (width*y)] = e_z;
					// Do Color
					Color colorWithLight; // The final color for this pixel (will/shall be)

					MipMapTextureMagicGo(colorWithLight, scene, T_terpolate, obj, dTdx, ActiveEdgeList[m].dTdy);

					Vector3D tempNterp = N_terpolate; // make a copy of N_terp.
					// Now, manip N_terpolate [bump it] for use in phong shading and bump maps.


					BumpTheNormal(N_terpolate, scene, T_terpolate, obj, dTdx, ActiveEdgeList[m].dTdy);

					/**********
					[just me testing stuff out, ignore]
					if (rand() % 5 == 2)
					{
						N_terpolate[0] = N_terpolate[0] + (rand()%6 * .1f);
						N_terpolate[1] = N_terpolate[1] + (rand()%6 * .1f);
						N_terpolate[2] = N_terpolate[2] + (rand()%6 * .1f);
					}
					**********/


					N_terpolate.Normalize(); // probably quite important yes?

					PhongShading(colorWithLight, colorWithLight, scene, N_terpolate, (float)x, (float)y, e_z, width, height, obj);

					N_terpolate = tempNterp; // restore N_terp to its glory [realness]

					// ----------------------------
					/*******************************************
					[ hack job way, was for testing]
					Vector3D Nn = N_terpolate;
					// They all have the same normal, so it is all good.
					Vector3D Pn = scene.lights[0].P; // just taking 1st existing light [only 1 in prj 1]
					Nn.Normalize();
					Pn.Normalize();

					const float dp = Dot (Nn, Pn);  // Old Note to self: check J vs K

					Color colorWithLight = color;
					colorWithLight.rgba[0] = color[0] * dp;
					colorWithLight.rgba[1] = color[1] * dp;
					colorWithLight.rgba[2] = color[2] * dp;
					colorWithLight.rgba[3] = color[3] * dp;


					// ----------------------------
					******************************************/
					colorBuffer[x + (width*y)] = colorWithLight;

				}
				e_z = e_z + dzdx; // increase zed as we walk along...
				N_terpolate = N_terpolate + dNdx; // neo
				T_terpolate = T_terpolate + dTdx; // neo2
			}
		} // m loop.
		



	} // end of 'y =' loop.




	EdgeList.clear();
	ActiveEdgeList.clear();
} // end of function.

////////////////////////////////////////////////////////////////////////
// This is called once after the scene is created but before it is
// ever draw to allow for any pre-processing steps.
void PreprocessScene(Scene& scene)
{

}

////////////////////////////////////////////////////////////////////////
// Procedure DrawScene is called whenever the scene needs to be drawn.
void DrawScene(Scene& scene, int width, int height)
{
	// Choose OpenGL or student rendering here.  The decision can be
	// based on useOpenGLRendering, useFastRendering, or both:

	// -----------------------------
	// Okay.  Think for a second.  We do NOT know what width and height
	// are until runtime.  These determine size of our lovely z-buffer.
	// So, dynamically allocate.

	std::vector<EdgeNode> EdgeList;
	assert(width >= 0);
	assert(height >= 0);

	// Okay, it is too expensive/slow to re-allocate memory each time, especially
	// if it is not necessary.  Therefore, allocate once, and don't reallocate again
	// unless height and/or width changes from previous call.

	if (g_prevHeight != height || g_prevWidth != width)
	{
		// Blow up existing zBuffer and colorBuffer and
		// reallocate them.
		delete [] zBuffer;  // Destroyed!
		delete [] colorBuffer;  // Destroyed!
		zBuffer = new float[width * height + 1000];  // Allocated!
		colorBuffer = new Color[width * height + 1000];  // Allocated!
		// I am making my buffers be larger by a size of 1000 to build in protection
		// in event of accidental writing out of bounds of array.  Better safe than sorry.
	}
	else
	{
		// Leave z-buffer allocation/size alone.
		// But "one" it out either way.
	}

	g_prevHeight = height;
	g_prevWidth = width;

	// Set Z-Buffer to one everywhere AND set color-buffer to background color [white].
	for (int i = 0; i < width * height; ++i)
	{
		zBuffer[i] = 1.f; // abuse of theory, I plan to treat as a 2D array (later)
		colorBuffer[i].rgba[0] = 1.f;
		colorBuffer[i].rgba[1] = 1.f;
		colorBuffer[i].rgba[2] = 1.f;
		colorBuffer[i].rgba[3] = 1.f;
	}

	// -----------------------------

	// if (useOpenGLRendering || useFastRendering)  {
	if (useOpenGLRendering)  {
		DrawSceneWithOpenGL(scene, width, height);
	}

	// Student rendering code goes here
	else		
	{
		// For Each Object O
			// For Each Polygon P
		
		// Wow, the Readme dot txt file has code to do this part, liberate it!


		// neo2
		// If we have not yet made MipMaps, make em for each obj's TM.
		// Only done ONCE EVER!
		if (gMIPMAPS.size() == 0)
		{
			// Make MM's for ALL TM's!
			for (int k=0;  k<(int)scene.objects.size();  k++) // for each object
			{
				Object& obj = scene.objects[k];
				assert (obj.texture != NULL);
				gMIPMAPS.push_back( MipMap(obj.texture->width, obj.texture->height, obj.texture) );
			}
		}
		// neo2


        for (int k=0;  k<(int)scene.objects.size();  k++) // for each object
		{
            Object& obj = scene.objects[k];
            for (int j=0;  j<(int)obj.polygons.size();  j++) // for each polygon
			{
                APolygon& poly = obj.polygons[j]; // [grabbed a polygon, use it!]
				////const float dp = Dot (poly[j].N, scene.lights[0].P);  // note: check J vs K
				////Color colorWithLight = obj.diffuse;
				////colorWithLight.rgba[0] = colorWithLight[0] * dp;
				////colorWithLight.rgba[1] = colorWithLight[1] * dp;
				////colorWithLight.rgba[2] = colorWithLight[2] * dp;
				////colorWithLight.rgba[3] = colorWithLight[3] * dp;


				/*************
				[ proj 1 coloring hackjab ]
				Vector3D Nn = poly[0].N; // Note to Self: Zero here in [=] works wonders!
				// They all have the same normal, so it is all good.
				Vector3D Pn = scene.lights[0].P; // just taking 1st existing light [only 1 in prj 1]
				Nn.Normalize();
				Pn.Normalize();

				const float dp = Dot (Nn, Pn);  // Old Note to self: check J vs K

				Color colorWithLight = obj.diffuse;
				colorWithLight.rgba[0] = obj.diffuse[0] * dp;
				colorWithLight.rgba[1] = obj.diffuse[1] * dp;
				colorWithLight.rgba[2] = obj.diffuse[2] * dp;
				colorWithLight.rgba[3] = obj.diffuse[3] * dp;
				************/

                // Do something with poly [-roger-]
				// Transform the polygon into Pixel Space.
				// Okay, think again!  APolygon is a std vec of Verticies!
				// We just made a reference to a vec of verticies making up a polygon.
				// Poly is a (ref to a) std vec of all verticies making 
				// up one of many polys which make up an obj.

				// OKAY!  Think some more!  We have all the verticies making
				// up a polygon, but we need these in pixel space my friend.
				// So, look/loop through ALL of the verticies of said polygon
				// and transform them into pixel-land.

				// But poly is a reference, so, don't manip contents of poly itself.
				// Why not make a deep copy and then manipulate the copy!

				/*****************
				[ from ReadMe File]
				The code to transform a world coordinate Vertex4D W to screen
				coordinates, then to pixel coordinates will look like this:
					Vector4D T = scene.viewing.Transform(W);
					Vector3D S = scene.projection.Transform(T).Hdiv();
					S[0] = (S[0]+1.0)*width/2.0;
					S[1] = (S[1]+1.0)*height/2.0;

				******************/

				EdgeList.clear();
				for (int a = 0; a < (int)poly.size(); ++a)
				{
					int prev = poly[a].prevIndex;
					Vector4D P = poly[prev].V;
					Vector4D Q = poly[a].V;

					Vector4D T1 = scene.viewing.Transform(P);
					Vector3D S1 = scene.projection.Transform(T1).Hdiv();
					
					Vector4D T2 = scene.viewing.Transform(Q);
					Vector3D S2 = scene.projection.Transform(T2).Hdiv();

					S1[0] = (S1[0]+1.0)*width/2.0;
					S1[1] = (S1[1]+1.0)*height/2.0;

					S2[0] = (S2[0]+1.0)*width/2.0;
					S2[1] = (S2[1]+1.0)*height/2.0;

					assert(T1[3] == 1); // just testing, always seems to be. neo2
					assert(T2[3] == 1); // just testing, always seems to be. neo2

					// WRONG: float zz1 = S1[2]; // neo2 [should be our 'w', a w val anyway]
					// WRONG:  float zz2 = S2[2]; // neo2 [should be our 'w', a w val anyway]

					// neo2 //
					Vector4D vtemp1 = scene.projection.Transform(T1); // don't do perspective proj yet.
					Vector4D vtemp2 = scene.projection.Transform(T2); // don't do perspective proj yet.
					float zz1 = vtemp1[3]; // just grabbed a 'w'
					float zz2 = vtemp2[3]; // just grabbed a 'w'
					// neo2 //

					Vertex vs1; // Vertex 1 of 2 of an Edge
					Vertex vs2; // Vertex 2 of 2 of an Edge
					vs1.V.xyzw[0] = S1[0];
					vs1.V.xyzw[1] = S1[1];
					vs1.V.xyzw[2] = S1[2];
					vs2.V.xyzw[0] = S2[0];
					vs2.V.xyzw[1] = S2[1];
					vs2.V.xyzw[2] = S2[2];

					{
						// Getting Vertex Normal.  Do NOT take to pixel space my friend.\
						// neo
						int prev = poly[a].prevIndex;
						vs1.N = poly[prev].N;
						vs2.N = poly[a].N;
						vs1.u = poly[prev].u; // neo2
						vs1.v = poly[prev].v; // neo2
						vs2.v = poly[a].v; // neo2
						vs2.u = poly[a].u; // neo2
					}

					// Make a new EdgeNode and append to list.
					// "May" remove it right away if edge is 'no good'
					EdgeList.push_back( EdgeNode(vs1, vs2, obj, zz1, zz2) );  // neo2 added params

					if ((int)EdgeList.back().y0 == (int)EdgeList.back().y1)
					{
						// Horizontal implies death
						EdgeList.pop_back(); // Sayonara
						continue;
					}

					// Note to self, less than OR equal to zero, make sure this is okay.
					if ( /*EdgeList.size() > 0 && */ EdgeList.back().y1 <= 0.f) // if ending of edge is ALSO off-screen.
					{
						// Not visible implies death
						EdgeList.pop_back(); // Sayonara
						continue;
					}

				}

				// Okay, before calling this, make sure my EdgeList is proper and
				// this will do the rest!
				// It is!!
				ScanLine(  EdgeList,  /*obj.diffuse*/ /*colorWithLight*/ width, height , scene, obj );
				EdgeList.clear(); // Empty for next time use.


			}

		}

		// Alright, now that I am here, I have filled in spots for the entire color
		// buffer.  So, lets go through the color-buffer and tell OpenGL to draw stuff.

         glBegin(GL_POINTS);   // Put OpenGL in point drawing mode
         for (int y = 0; y < height; ++y)
		 {
			 for (int x = 0; x < width; ++x)
			 {
				//glColor3f(r,g,b); // Set color for next pixel
				 glColor3f( colorBuffer[x + (width*y)].rgba[0], colorBuffer[x + (width*y)].rgba[1], colorBuffer[x + (width*y)].rgba[2]   ); // Set color for next pixel
				glVertex2i(x,y);  // Put that color at pixel (x,y)
			 }
		 }
         glEnd();   

	} // <--- end of the mega super awesome special else case;

	
} // <--- end of function

